#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime
import logging
from typing import Optional

from th2_common_utils import create_event, create_event_id, message_to_table
from th2_grpc_common.common_pb2 import Event, EventID, EventStatus, Message

logger = logging.getLogger()


def create_grpc_method_event(request_event_id: EventID,
                             method_name: str,
                             description: Optional[str]) -> Event:
    if not description:
        description = f'{method_name} - {datetime.now()}'

    event: Event = create_event(event_id=create_event_id(),
                                parent_id=request_event_id,
                                name=description,
                                event_type=method_name)
    return event


def create_responses_root_event(method_event_id: EventID,
                                status: int,
                                event_name: str) -> Event:
    event: Event = create_event(event_id=create_event_id(),
                                parent_id=method_event_id,
                                status=status,
                                name=event_name,
                                event_type='Incoming messages')
    return event


def create_send_request_succeed_event(message: Message, act_event_id: EventID) -> Event:
    event: Event = create_event(event_id=create_event_id(),
                                parent_id=act_event_id,
                                name=f"Send '{message.metadata.message_type}' message from Act",
                                event_type='Outgoing message',
                                body=_convert_message_to_event_body(message))
    return event


def create_receive_response_succeed_event(response: Message,
                                          responses_root_event: EventID,
                                          status: int) -> Event:
    event: Event = create_event(event_id=create_event_id(),
                                parent_id=responses_root_event,
                                event_type='Incoming message',
                                status=status,
                                name=response.metadata.message_type,
                                body=_convert_message_to_event_body(response),
                                attached_message_ids=[response.metadata.id])
    return event


def create_receive_response_failed_event(text: str, responses_root_event: EventID) -> Event:
    event: Event = create_event(event_id=create_event_id(),
                                parent_id=responses_root_event,
                                event_type='Act message',
                                status=EventStatus.FAILED,
                                name=text)
    return event


def create_context_cancelled_event(method_event_id: EventID) -> Event:
    event: Event = create_event(event_id=create_event_id(),
                                parent_id=method_event_id,
                                event_type='gRPC context failure',
                                status=EventStatus.FAILED,
                                name='gRPC context was cancelled by client')
    return event


def _convert_message_to_event_body(message: Message) -> bytes:
    if isinstance(message, Message):
        message_event_body = message_to_table(message)
        return bytes(message_event_body)

    return b''



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import importlib
import logging
from pathlib import Path
import traceback
from typing import Any, Callable, Dict, Optional

from th2_act.handler_attributes import HandlerAttributes
from th2_act.util.subscription_manager import SubscriptionManager
from th2_common.schema.grpc.router.grpc_router import GrpcRouter
from th2_common.schema.message.message_router import MessageRouter
from th2_grpc_check1.check1_service import Check1Service

logger = logging.getLogger()


class Act:
    """Contains initialized ActHandler classes (uploaded from files) instances.

    Args:
        grpc_router (GrpcRouter): gRPC router from CommonFactory.
        message_router (MessageRouter): Message router from CommonFactory.
        event_router (MessageRouter): Event router from CommonFactory.
        handlers_package_relative_path (str, optional): A relative path to the package with the handlers.
            Defaults to '/handlers'.
    """

    def __init__(self,
                 grpc_router: GrpcRouter,
                 message_router: MessageRouter,
                 event_router: MessageRouter,
                 handlers_package_relative_path: str = 'handlers'):
        subscription_manager = SubscriptionManager()
        message_router.subscribe_all(subscription_manager)

        self._handler_attrs = HandlerAttributes(check1_connector=grpc_router.get_service(Check1Service),  # type: ignore
                                                message_router=message_router,
                                                event_router=event_router,
                                                subscription_manager=subscription_manager)

        self.handlers = self._load_handlers(handlers_package_relative_path)

    def _load_handlers(self, handlers_package_relative_path: str) -> Dict[Callable, Callable]:
        """Uploads ActHandler classes from files and initialize them. Returns a dict with ActHandler class instance
        as key and add_servicer_to_server() function as value. The dict can be passed as an argument when
        initializing an instance of the ActServer class.
        """

        handlers_servicers_dict = {}

        handlers_package = importlib.import_module(handlers_package_relative_path)
        handlers_package_name = handlers_package.__name__
        handlers_paths = Path(handlers_package.__path__[0]).glob('*.py')

        for handler_path in handlers_paths:
            try:
                handler_class = _get_handler_class_from_module(handlers_package_name, handler_path)

                if handler_class:
                    handler = handler_class(self._handler_attrs)
                    handlers_servicers_dict[handler] = _get_func_add_servicer_to_server(handler)
            except Exception as e:
                logger.error(f'Cannot upload handler with the path '
                             f'{handlers_package.__name__}.{handler_path.stem}: {e}'
                             f'\n{"".join(traceback.format_tb(e.__traceback__))}')

        logger.info(f'Handlers upload finished. Total: {len(handlers_servicers_dict)} handler(s)')

        return handlers_servicers_dict


def _get_handler_class_from_module(handlers_package_name: str, handler_path: Path) -> Optional[Callable]:
    handler_module = importlib.import_module(f'{handlers_package_name}.{handler_path.stem}')

    return getattr(handler_module, 'ActHandler', None)  # type: ignore


def _get_func_add_servicer_to_server(handler: Any) -> Any:
    servicer_class = handler.__class__.__base__
    servicer_module = importlib.import_module(servicer_class.__module__)

    return getattr(servicer_module, f'add_{servicer_class.__name__}_to_server')



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from typing import Dict, List, Optional, Union

from google.protobuf.text_format import MessageToString
from th2_act.act_response import ActMultiResponse, ActResponse
from th2_act.util.act_sender import ActSender
from th2_grpc_common.common_pb2 import EventID, Message, RequestStatus

logger = logging.getLogger()


class RequestProcessorUtils:

    ResponsesType = Union[List[ActResponse], ActMultiResponse]

    def __init__(self,
                 act_sender: ActSender):
        self.act_sender = act_sender
        self.checkpoint = act_sender.checkpoint

    def _create_and_send_responses_root_event(self,
                                              receive_method_name: str,
                                              messages_received: int = 0,
                                              messages_expected: Optional[int] = None) -> EventID:
        if messages_expected in (1, None):
            return self.act_sender.create_and_send_responses_root_event(status=RequestStatus.SUCCESS,
                                                                        event_name=f'Received {messages_received} '
                                                                                   f'message(s) - '
                                                                                   f'{receive_method_name}')
        elif messages_received == messages_expected:
            return self.act_sender.create_and_send_responses_root_event(status=RequestStatus.SUCCESS,
                                                                        event_name=f'Received {messages_received}/'
                                                                                   f'{messages_expected} message(s)'
                                                                                   f' - {receive_method_name}')
        else:
            return self.act_sender.create_and_send_responses_root_event(status=RequestStatus.ERROR,
                                                                        event_name=f'Received {messages_received}/'
                                                                                   f'{messages_expected} message(s)'
                                                                                   f' - {receive_method_name}')

    def create_and_send_responses_events(self,
                                         receive_method_name: str,
                                         runtime: float,
                                         act_responses: Optional[ResponsesType],
                                         number_of_responses_expected: Optional[int] = None) -> None:
        if act_responses is None:
            responses_root_event = self._create_and_send_responses_root_event(
                receive_method_name=receive_method_name,
                messages_expected=number_of_responses_expected
            )
            error_text = f'Message matching the condition was not found in {round(runtime, 1)} s'
            self.act_sender.create_and_send_response_failed_event(text=error_text,
                                                                  responses_root_event=responses_root_event)

        else:
            responses_root_event = self._create_and_send_responses_root_event(
                receive_method_name=receive_method_name,
                messages_received=len(act_responses),
                messages_expected=number_of_responses_expected
            )

            if isinstance(act_responses, list):
                for response in act_responses:
                    self.act_sender.create_and_send_response_succeed_event(response=response.message,  # type: ignore
                                                                           responses_root_event=responses_root_event,
                                                                           status=response.status.status)

                logger.debug(
                    f'Received {len(act_responses)} messages: '  # type: ignore
                    f'{[MessageToString(act_response.message, as_one_line=True) for act_response in act_responses]}'
                )

            elif isinstance(act_responses, ActMultiResponse):
                for message in act_responses.messages:
                    self.act_sender.create_and_send_response_succeed_event(response=message,
                                                                           responses_root_event=responses_root_event,
                                                                           status=act_responses.status.status)

                logger.debug(
                    f'Received {len(act_responses)} messages: '
                    f'{[MessageToString(m, as_one_line=True) for m in act_responses.messages]}'
                )

    def create_act_responses_list(self,
                                  status_messages_dict: Dict[int, List[Message]]) -> Optional[List[ActResponse]]:
        if status_messages_dict:
            return [
                ActResponse(message=message, status=status, checkpoint=self.checkpoint)
                for status, messages in status_messages_dict.items()
                for message in messages
            ]

        return None

    def create_act_multi_response(self,
                                  status_messages_dict: Dict[int, List[Message]]) -> Optional[ActMultiResponse]:
        for status, messages in status_messages_dict.items():
            return ActMultiResponse(messages=messages,
                                    status=status,
                                    checkpoint=self.checkpoint)

        return None



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Iterator, List, Optional, Union

from th2_grpc_common.common_pb2 import Checkpoint, Message, RequestStatus


class ActResponse:
    """ActResponse may be returned to act implementation by Act. The class instance can contain message, status and
    checkpoint fields, which can be used when creating custom response objects.
    """

    __slots__ = ('message', 'status', 'checkpoint')

    def __init__(self,
                 message: Optional[Message] = None,
                 status: Union[str, int] = RequestStatus.SUCCESS,
                 checkpoint: Optional[Checkpoint] = None) -> None:
        self.message = message
        self.status = RequestStatus(status=status)  # type: ignore
        self.checkpoint = checkpoint


class ActMultiResponse:
    """ActMultiResponse may be returned to act implementation by Act. The class instance can contain list of messages,
    status and checkpoint fields, which can be used when creating custom response objects.
    """

    __slots__ = ('messages', 'status', 'checkpoint')

    def __init__(self,
                 messages: Optional[List[Message]] = None,
                 status: Union[str, int] = RequestStatus.SUCCESS,
                 checkpoint: Optional[Checkpoint] = None) -> None:
        if messages is None:
            self.messages = []
        else:
            self.messages = messages
        self.status = RequestStatus(status=status)  # type: ignore
        self.checkpoint = checkpoint

    def __iter__(self) -> Iterator:
        return self.messages.__iter__()

    def __getitem__(self, item: int) -> Message:
        return self.messages[item]

    def __len__(self) -> int:
        return len(self.messages)



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import defaultdict
from itertools import islice
import logging
from time import time
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Union

from th2_act.util.act_receiver import check
from th2_act.util.cache import Cache
from th2_act.util.grpc_context_manager import GRPCContextManager
from th2_grpc_common.common_pb2 import Message

logger = logging.getLogger()


StatusMessagesDict = Dict[int, List[Message]]


class CacheFilter:
    """Contains methods to get messages from the cache."""

    def __init__(self, context_manager: GRPCContextManager, cache: Cache) -> None:
        self.context_manager = context_manager

        self.cache: Cache = cache
        self.last_filtered_message_index: Optional[int] = None

    def get_n_matching_within_timeout(self,
                                      message_filters: Dict[Callable, int],
                                      n: int,
                                      check_previous_messages: bool,
                                      timeout: Union[int, float],
                                      start_time: float) -> Tuple[StatusMessagesDict, Optional[int]]:
        """Returns first N received messages that match the message_filters from cache.

        The method will check cache until all N messages are found or until context is no more active
        or until timeout is over.
        """

        status_messages_dict: Dict[int, List[Message]] = defaultdict(list)
        matching_message_index: Optional[int] = None
        start_index = self._get_start_index(check_previous_messages)
        messages_to_receive = n

        while messages_to_receive > 0:
            if self.context_manager.is_context_active() and time() - start_time < timeout:
                stop_index = self.cache.size

                status_message_iterator = self.get_filtered_message_iterator(message_filters=message_filters,
                                                                             messages_to_receive=messages_to_receive,
                                                                             start_index=start_index,
                                                                             stop_index=stop_index)

                for index, status, message in status_message_iterator:
                    status_messages_dict[status].append(message)
                    matching_message_index = index
                    messages_to_receive -= 1

                start_index = stop_index
            else:
                break

        self._update_last_filtered_message_index(index=matching_message_index)

        return status_messages_dict, matching_message_index

    def get_n_matching_immediately(self,
                                   message_filters: Dict[Callable, int],
                                   check_previous_messages: bool,
                                   n: Optional[int] = None) -> Tuple[StatusMessagesDict, Optional[int]]:
        """Returns all received messages that match the message_filter from cache.

        The method will check cache only once.
        """

        status_messages_dict: Dict[int, List[Message]] = defaultdict(list)
        matching_message_index: Optional[int] = None
        start_index = self._get_start_index(check_previous_messages)

        if self.context_manager.is_context_active():
            status_message_iterator = self.get_filtered_message_iterator(message_filters=message_filters,
                                                                         messages_to_receive=n,
                                                                         start_index=start_index,
                                                                         stop_index=self.cache.size)
            for index, status, message in status_message_iterator:
                status_messages_dict[status].append(message)
                matching_message_index = index

        self._update_last_filtered_message_index(index=matching_message_index)

        return status_messages_dict, matching_message_index

    def get_filtered_message_iterator(self,
                                      message_filters: Dict[Callable, int],
                                      messages_to_receive: Optional[int],
                                      start_index: int,
                                      stop_index: int) -> Iterator[Tuple[int, int, Message]]:
        return islice(
            (
                (index, status, message)
                for index, message in islice(enumerate(self.cache), start_index, stop_index)
                for condition, status in message_filters.items()
                if check(condition, message)
            ),
            0,
            messages_to_receive
        )

    def get_all_before_matching_from_cache(self, index: Optional[int]) -> List[Message]:
        if index is not None:
            return self.cache[0:index + 1]  # type: ignore
        else:
            return []

    def get_all_after_matching_from_cache(self, index: Optional[int]) -> List[Message]:
        if index is not None:
            return self.cache[index:self.cache.size]  # type: ignore
        else:
            return []

    def _get_start_index(self, check_previous_messages: bool) -> int:
        if check_previous_messages or self.last_filtered_message_index is None:
            return 0
        else:
            return self.last_filtered_message_index

    def _update_last_filtered_message_index(self, index: Optional[int]) -> None:
        if index is not None:
            self.last_filtered_message_index = index



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import threading
from typing import Union

from grpc import RpcContext

logger = logging.getLogger()


class GRPCContextManager:

    def __init__(self, context: RpcContext):
        self.context = context

        self.time_lock = threading.RLock()
        self.time_condition = threading.Condition(self.time_lock)

    def context_time_remaining(self) -> float:
        return self.context.time_remaining()  # type: ignore

    def is_context_active(self) -> bool:
        return self.context.is_active()  # type: ignore

    def cancel_context(self) -> None:
        self.context.cancel()

    def wait(self, seconds: Union[int, float]) -> None:
        if self.context.time_remaining() > seconds:
            self._await_sync(seconds)
        else:
            logger.warning('Remaining time of context less than wait time. Cache check starts now.')

    def _await_sync(self, timeout: Union[int, float]) -> None:
        try:
            self.time_lock.acquire()
            self.time_condition.wait(timeout)
        finally:
            self.time_lock.release()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from time import time
import traceback
from typing import Any, Callable, Dict, List, Optional, Union

from th2_act.act_response import ActMultiResponse, ActResponse
from th2_act.grpc_method_attributes import GrpcMethodAttributes
from th2_act.handler_attributes import HandlerAttributes
from th2_act.util.act_receiver import ActReceiver
from th2_act.util.act_sender import ActSender
from th2_act.util.cache import Cache
from th2_act.util.cache_filter import CacheFilter
from th2_act.util.grpc_context_manager import GRPCContextManager
from th2_act.util.request_processor_utils import RequestProcessorUtils
from th2_grpc_common.common_pb2 import Direction, Message, RequestStatus

logger = logging.getLogger()


class RequestProcessor:
    """A context manager used for sending requests and filtering messages from system under test.

    Mainly used when implementing gRPC methods in ActHandler class.

    Args:
        handler_attrs (HandlerAttributes): HandlerAttributes class instance. Generates by Act when loading handlers.
        method_attrs (GrpcMethodAttributes): GrpcMethodAttributes class instance. Initialize it with gRPC context,
            request event ID, gRPC method name and request description.
        prefilter (lambda or function, optional): Filter for messages that will be placed in cache.
            Defaults to None - cache will collect all messages.
    """

    def __init__(self,
                 handler_attrs: HandlerAttributes,
                 method_attrs: GrpcMethodAttributes,
                 prefilter: Optional[Callable] = None):
        self._subscription_manager = handler_attrs.subscription_manager
        self._grpc_context_manager = GRPCContextManager(context=method_attrs.context)

        self._cache: Cache = Cache()
        self._cache_filter: CacheFilter = CacheFilter(context_manager=self._grpc_context_manager, cache=self._cache)
        self._act_receiver: ActReceiver = ActReceiver(prefilter=prefilter, cache=self._cache)

        self._act_sender = ActSender(message_router=handler_attrs.message_router,
                                     event_router=handler_attrs.event_router,
                                     grpc_context_manager=self._grpc_context_manager,
                                     check1_connector=handler_attrs.check1_connector,
                                     method_attrs=method_attrs)
        self._checkpoint = self._act_sender.checkpoint

        self._rp_utils = RequestProcessorUtils(act_sender=self._act_sender)

    def __enter__(self) -> Any:
        self._subscription_manager.register(message_listener=self._act_receiver.message_listener)
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Optional[bool]:
        try:
            self._subscription_manager.unregister(message_listener=self._act_receiver.message_listener)
            self.clear_cache()
            if exc_type:
                logger.error(f'Exception occurred in RequestProcessor: {exc_type}'
                             f'\n{"".join(traceback.format_tb(exc_tb))}')
                return True
            return None
        except Exception as e:
            logger.error(f'Could not stop subscriber monitor: {e}\n{"".join(traceback.format_tb(e.__traceback__))}')
            return None

    def send(self,
             message: Message,
             echo_key_field: Optional[str] = None,
             timeout: Optional[Union[int, float]] = None) -> Optional[str]:
        """Sends message to conn. Returns 'MsgSeqNum' field of sent message's echo if 'echo_key_field' is set,
        else None.

        Args:
            message (Message): Message to send.
            echo_key_field (str): Key field to catch sent message's echo. If not set, no echo will be returned.
            timeout (int or float, optional): The time (in seconds) during which you are ready to wait for the echo.
                The echo will be returned as soon as it is found, without waiting for the end of the timeout.
                If not set, the gRPC context time will be used.

        Returns:
            'MsgSeqNum' if 'echo_key_field' is set, else None.
        """
        if not self._act_sender.send_message(message):
            self._grpc_context_manager.cancel_context()
            logger.info('Context was cancelled because of message sending fail')
            return None

        if echo_key_field is not None:
            echo_message = self._receive_request_message_echo(message, echo_key_field, timeout)
            if echo_message is not None:
                return echo_message.fields['header'].message_value.fields['MsgSeqNum'].simple_value

        return None

    def receive_first_matching(self,
                               message_filters: Dict[Callable, int],
                               timeout: Optional[Union[int, float]] = None,
                               check_previous_messages: bool = True) -> ActResponse:
        """Returns the first message (wrapped in ActResponse class) from cache that matches one of the filters.
        If no message found, None will be returned.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            timeout (int or float, optional): The time (in seconds) during which you are ready to wait for the message.
                The message will be returned as soon as it is found, without waiting for the end of the timeout.
                If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            ActResponse class instance with received message, status and checkpoint, if any message found.
        """

        start_time = time()

        if timeout is None:
            timeout = self._grpc_context_manager.context_time_remaining()

        status_messages_dict, _ = self._cache_filter.get_n_matching_within_timeout(
            message_filters=message_filters,
            n=1,
            check_previous_messages=check_previous_messages,
            timeout=timeout,
            start_time=start_time
        )

        runtime = time() - start_time

        act_responses = self._rp_utils.create_act_responses_list(status_messages_dict=status_messages_dict)
        self._rp_utils.create_and_send_responses_events(receive_method_name='first matching',
                                                        runtime=runtime,
                                                        act_responses=act_responses)

        if act_responses is not None:
            return act_responses[0]

        return ActResponse(status=RequestStatus.ERROR, checkpoint=self._checkpoint)

    def receive_first_n_matching(self,
                                 message_filters: Dict[Callable, int],
                                 n: int,
                                 timeout: Optional[Union[int, float]] = None,
                                 check_previous_messages: bool = True) -> List[ActResponse]:
        """Returns a list of first N messages (each wrapped in ActResponse class) from cache that match
        the filters. If no messages found, None will be returned.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            n (int): A number of filtered messages you want to receive.
            timeout (float, optional): The time (in seconds) during which you are ready to wait for the messages.
                The messages will be returned as soon as N messages are found, without waiting for the end of the
                timeout. If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            List of ActResponse class instances with received message, status and checkpoint, if any message found.
        """

        start_time = time()

        if timeout is None:
            timeout = self._grpc_context_manager.context_time_remaining()

        status_messages_dict, _ = self._cache_filter.get_n_matching_within_timeout(
            message_filters=message_filters,
            n=n,
            check_previous_messages=check_previous_messages,
            timeout=timeout,
            start_time=start_time
        )

        runtime = time() - start_time

        act_responses = self._rp_utils.create_act_responses_list(status_messages_dict=status_messages_dict)
        self._rp_utils.create_and_send_responses_events(receive_method_name='first n matching',
                                                        act_responses=act_responses,
                                                        number_of_responses_expected=n,
                                                        runtime=runtime)

        return act_responses or [ActResponse(status=RequestStatus.ERROR, checkpoint=self._checkpoint)]

    def receive_all_before_matching(self,
                                    message_filters: Dict[Callable, int],
                                    timeout: Optional[Union[int, float]] = None,
                                    check_previous_messages: bool = True) -> ActMultiResponse:
        """Returns an ActMultiResponse with messages from cache that precede a message that
        matches one of the filters. Message that matching condition is included in the list.
        If no messages found, None will be returned.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            timeout (float, optional): The time (in seconds) during which you are ready to wait for the messages.
                The messages will be returned as soon as the first matching message is found,
                without waiting for the end of the timeout. If not set, the gRPC context time will be used.
            check_previous_messages (bool, optional): Set True if you want to look for the messages throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            List of ActResponse class instances with received message, status and checkpoint, if any message found.
        """

        start_time = time()

        if timeout is None:
            timeout = self._grpc_context_manager.context_time_remaining()

        status_messages_dict, index = self._cache_filter.get_n_matching_within_timeout(
            message_filters=message_filters,
            n=1,
            check_previous_messages=check_previous_messages,
            timeout=timeout,
            start_time=start_time
        )

        runtime = time() - start_time

        for status in status_messages_dict:
            status_messages_dict[status] = self._cache_filter.get_all_before_matching_from_cache(index=index)
        act_multi_response = self._rp_utils.create_act_multi_response(status_messages_dict=status_messages_dict)
        self._rp_utils.create_and_send_responses_events(receive_method_name='all before matching',
                                                        act_responses=act_multi_response,
                                                        runtime=runtime)

        return act_multi_response or ActMultiResponse(status=RequestStatus.ERROR, checkpoint=self._checkpoint)

    def receive_all_after_matching(self,
                                   message_filters: Dict[Callable, int],
                                   wait_time: Optional[Union[int, float]] = None,
                                   check_previous_messages: bool = True) -> ActMultiResponse:
        """Returns a ActMultiResponse with messages from cache that follows a message that
        matches one of the filters. Message that matching condition is included in the list.
        If no messages found, None will be returned.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            wait_time (int, optional): The time (in seconds) that the Act will wait before checking the cache for
                the matching message. Defaults to None (cache will be checked immediately and only once).
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            List of ActResponse class instances with received message, status and checkpoint, if any message found.
       """

        start_time = time()

        if wait_time:
            self._grpc_context_manager.wait(wait_time)

        status_messages_dict, index = self._cache_filter.get_n_matching_immediately(
            message_filters=message_filters,
            check_previous_messages=check_previous_messages,
            n=1
        )

        runtime = time() - start_time

        for status in status_messages_dict:
            status_messages_dict[status] = self._cache_filter.get_all_after_matching_from_cache(index=index)
        act_multi_response = self._rp_utils.create_act_multi_response(status_messages_dict=status_messages_dict)
        self._rp_utils.create_and_send_responses_events(receive_method_name='all after matching',
                                                        act_responses=act_multi_response,
                                                        runtime=runtime)

        return act_multi_response or ActMultiResponse(status=RequestStatus.ERROR, checkpoint=self._checkpoint)

    def receive_all_matching(self,
                             message_filters: Dict[Callable, int],
                             wait_time: Optional[Union[int, float]] = None,
                             check_previous_messages: bool = True) -> List[ActResponse]:
        """Returns a list of messages (wrapped in ActResponse class) from cache that matches filters.
        If no messages found, None will be returned.

        Args:
            message_filters (Dict[lambda, RequestStatus]): The filter (lambda or function) through which the messages
                from cache will be filtered out - as key, RequestStatus (SUCCESS or ERROR) - as value.
            wait_time (int, optional): The time that the Act will wait before checking the cache for matching messages.
                Defaults to None (cache will be checked immediately).
            check_previous_messages (bool, optional): Set True if you want to look for the message throughout all
                received messages; set False if you want to check only newly received messages. Defaults to True.

        Returns:
            List of ActResponse class instances with received message, status and checkpoint, if any message found.
        """

        start_time = time()

        if wait_time:
            self._grpc_context_manager.wait(wait_time)

        status_messages_dict, index = self._cache_filter.get_n_matching_immediately(
            message_filters=message_filters,
            check_previous_messages=check_previous_messages
        )

        runtime = time() - start_time

        act_responses = self._rp_utils.create_act_responses_list(status_messages_dict=status_messages_dict)
        self._rp_utils.create_and_send_responses_events(receive_method_name='all matching',
                                                        act_responses=act_responses,
                                                        runtime=runtime)

        return act_responses or [ActResponse(status=RequestStatus.ERROR, checkpoint=self._checkpoint)]

    @property
    def cache_size(self) -> int:
        """Returns a number of messages in cache."""

        return self._cache.size

    def clear_cache(self) -> None:
        """Deletes all messages from cache."""

        self._cache.clear()

    def _receive_request_message_echo(self,
                                      message: Message,
                                      echo_key_field: str,
                                      timeout: Optional[Union[int, float]]) -> Optional[Message]:
        start_time = time()

        if timeout is None:
            timeout = self._grpc_context_manager.context_time_remaining()

        echo_filter = lambda echo: (  # noqa: E731, ECE001
                echo.fields[echo_key_field] == message.fields[echo_key_field]
                and echo.metadata.message_type == message.metadata.message_type
                and echo.metadata.id.direction == Direction.SECOND
                and echo.metadata.id.connection_id.session_alias == message.metadata.id.connection_id.session_alias
        )

        status_messages_dict, _ = self._cache_filter.get_n_matching_within_timeout(
            message_filters={echo_filter: RequestStatus.SUCCESS},
            n=1,
            check_previous_messages=False,
            timeout=timeout,
            start_time=start_time
        )

        act_responses = self._rp_utils.create_act_responses_list(status_messages_dict=status_messages_dict)

        if act_responses is not None:
            return act_responses[0].message

        logger.error(f'Cannot find echo with key field {echo_key_field} of message '
                     f'with type {message.metadata.message_type} and '
                     f'session alias {message.metadata.id.connection_id.session_alias}')
        return None



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from typing import Callable, Dict

import grpc

logger = logging.getLogger()


class ActServer:
    """Act server using ActHandler class instances described in act implementation.

    Args:
        server (grpc.Server): Server of gRPC router from CommonFactory.
        handlers (Dict[ActHandler, add_servicer_to_server()]): a dict with ActHandler class instance as key and
            add_servicer_to_server() function as value. This parameter can be created by Act.load_handlers() method.
    """

    def __init__(self, server: grpc.Server, handlers: Dict[Callable, Callable]) -> None:
        self.server = server
        self.handlers = handlers

        self.is_running = False

    def start(self) -> None:
        """Starts the server. Returns None"""

        if self.handlers:
            for handler in self.handlers:
                add_servicer_to_server_function = self.handlers[handler]
                add_servicer_to_server_function(handler, self.server)

            self.server.start()
            self.is_running = True
            logger.info('GRPC Server started')
            logger.info(f'Services: {[service.service_name() for service in self.server._state.generic_handlers]}')

            self.server.wait_for_termination()
        else:
            logger.error('Cannot start GRPC server: no handlers uploaded')

    def stop(self) -> None:
        """Stops the server. Returns None"""

        if self.is_running:
            self.server.stop(None)
        logger.info('GRPC Server stopped')



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import grpc
from th2_grpc_common.common_pb2 import EventID


class GrpcMethodAttributes:
    """Contains parameters of gRPC implementation method. GrpcMethodAttributes are used in RequestProcessor
    context manager.

    Args:
        context (grpc.RpcContext): gRPC method context.
        request_event_id (EventID): Request event ID.
        method_name (str): Act name. This will be shown in GIU as gRPC method event name.
        request_description (str, optional): Request description. This will be shown in GIU as request event name.
    """

    __slots__ = ('context', 'method_name', 'request_event_id', 'request_description')

    def __init__(self,
                 context: grpc.RpcContext,
                 method_name: str,
                 request_event_id: EventID,
                 request_description: str = ''):
        self.context = context
        self.method_name = method_name
        self.request_event_id = request_event_id
        self.request_description = request_description



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from typing import Dict, Iterable, Iterator, List, Union

from th2_grpc_common.common_pb2 import Message

logger = logging.getLogger()


StatusMessagesDict = Dict[int, List[Message]]


class Cache:
    """Stores a cache of received messages."""

    def __init__(self) -> None:
        self.cache: List[Message] = []

    def __iter__(self) -> Iterator:
        return self.cache.__iter__()

    def __getitem__(self, item: Union[int, slice]) -> Union[Message, List[Message]]:
        return self.cache[item]

    @property
    def size(self) -> int:
        return len(self.cache)

    def add(self, messages: Iterable[Message]) -> None:
        self.cache.extend(messages)

    def clear(self) -> None:
        self.cache.clear()



#   Copyright 2020-2022 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import traceback

from google.protobuf.text_format import MessageToString
from th2_act.grpc_method_attributes import GrpcMethodAttributes
import th2_act.util.act_events as events
from th2_act.util.grpc_context_manager import GRPCContextManager
from th2_common.schema.message.message_router import MessageRouter
from th2_grpc_check1.check1_pb2 import CheckpointRequest, CheckpointResponse
from th2_grpc_check1.check1_service import Check1Service
from th2_grpc_common.common_pb2 import Checkpoint, Event, EventBatch, EventID, Message, MessageBatch

logger = logging.getLogger()


class ActSender:
    """Sends messages and events."""

    def __init__(self,
                 message_router: MessageRouter,
                 event_router: MessageRouter,
                 check1_connector: Check1Service,
                 grpc_context_manager: GRPCContextManager,
                 method_attrs: GrpcMethodAttributes):
        self.message_router = message_router
        self.event_router = event_router
        self.check1_connector = check1_connector

        self.grpc_context_manager = grpc_context_manager
        self.method_name = method_attrs.method_name

        self.is_sending_allowed = True

        self.method_event_id = self.create_and_send_method_event(request_event_id=method_attrs.request_event_id,
                                                                 method_name=method_attrs.method_name,
                                                                 description=method_attrs.request_description)
        self.checkpoint = self.register_checkpoint()

    def register_checkpoint(self) -> Checkpoint:
        checkpoint = CheckpointRequest(parent_event_id=self.method_event_id)
        check1_response: CheckpointResponse = self.check1_connector.createCheckpoint(request=checkpoint)

        return check1_response.checkpoint

    def send_message(self, message: Message) -> bool:
        message.parent_event_id.CopyFrom(self.method_event_id)

        if self._send_with_message_router(message):
            event = events.create_send_request_succeed_event(message=message, act_event_id=self.method_event_id)
            self._send_with_event_router(event)

            logger.debug('Message was sent successfully: %s' % MessageToString(message, as_one_line=True))
            return True

        return False

    def create_and_send_method_event(self, request_event_id: EventID, method_name: str, description: str) -> EventID:
        event = events.create_grpc_method_event(request_event_id=request_event_id,
                                                method_name=method_name,
                                                description=description)
        self._send_with_event_router(event)

        return event.id

    def create_and_send_responses_root_event(self, status: int, event_name: str) -> EventID:
        event = events.create_responses_root_event(method_event_id=self.method_event_id,
                                                   status=status,
                                                   event_name=event_name)
        self._send_with_event_router(event)
        return event.id

    def create_and_send_response_succeed_event(self,
                                               response: Message,
                                               responses_root_event: EventID,
                                               status: int) -> EventID:
        event = events.create_receive_response_succeed_event(response=response,
                                                             responses_root_event=responses_root_event,
                                                             status=status)
        self._send_with_event_router(event)

        return event.id

    def create_and_send_response_failed_event(self, text: str, responses_root_event: EventID) -> EventID:
        event = events.create_receive_response_failed_event(text=text,
                                                            responses_root_event=responses_root_event)
        self._send_with_event_router(event)

        return event.id

    def _send_with_event_router(self, event: Event) -> None:
        try:
            if not self.is_sending_allowed:
                return
            elif self.grpc_context_manager.is_context_active():
                self.event_router.send_all(message=EventBatch(events=[event]))
            else:
                self.is_sending_allowed = False
                event = events.create_context_cancelled_event(self.method_event_id)
                self.event_router.send_all(message=EventBatch(events=[event]))
                logger.warning(f'Cannot send event: {self.method_name} gRPC context was cancelled by client')

        except Exception as e:
            logger.error(f'Cannot send event "{event.name}" with attached messages IDs '
                         f'{event.attached_message_ids} to estore: {e}'
                         f'\n{"".join(traceback.format_tb(e.__traceback__))}')

    def _send_with_message_router(self, message: Message) -> bool:
        try:
            if not self.is_sending_allowed:
                return False

            elif self.grpc_context_manager.is_context_active():
                self.message_router.send(MessageBatch(messages=[message]))
                return True

            else:
                self.is_sending_allowed = False
                event = events.create_context_cancelled_event(self.method_event_id)
                self.event_router.send_all(message=EventBatch(events=[event]))
                logger.warning(f'Cannot send message: {self.method_name} gRPC context was cancelled by client')
                return False

        except Exception as e:
            logger.error(f'Cannot send the message with session alias {e}'
                         f'{message.metadata.id.connection_id.session_alias}: '
                         f'\n{"".join(traceback.format_tb(e.__traceback__))}')
            return False
